require_relative 'db_connection'
require_relative 'associatable'
require_relative 'searchable'
require 'active_support/inflector'

module RailzLite
  class SQLObject
    extend RailzLite::Associatable
    extend RailzLite::Searchable

    def self.columns
      @columns ||= DBConnection.execute2(<<-SQL)
      SELECT
        *
      FROM
        #{table_name};
    SQL
      @columns.first.map(&:to_sym)
    end

    def self.finalize!
      columns.each do |name|
        define_method(name) do
          attributes[name]
        end
        define_method("#{name}=") do |val|
          attributes[name] = val
        end
      end
    end

    def self.table_name=(table_name)
      @table_name = table_name
    end

    def self.table_name
      @table_name || self.name.tableize 
    end

    def self.all
      results = DBConnection.execute(<<-SQL)
      SELECT
        *
      FROM
        #{table_name};
    SQL
      parse_all(results)
    end

    def self.parse_all(results)
      results.map { |attrs| self.new(attrs) }
    end

    def self.find(id)
      target = DBConnection.execute(<<-SQL, id)
      SELECT
        *
      FROM
        #{table_name}
      WHERE
        ID = ?;
    SQL

      return nil if target.empty?
      self.new(target.first)
    end

    # for insert statements we don't want to include the id field
    def self.columns_sans_id
      self.columns.reject { |col| col == :id }
    end

    def initialize(params = {})
      params.each do |attr_name, val|
        name_sym = attr_name.to_sym
        raise Exception.new "unknown attribute '#{attr_name}'" unless self.class.columns.include?(name_sym)
        send("#{name_sym}=", val)
      end
    end

    def attributes
      @attributes ||= {}
    end

    def attribute_values
      self.class.columns.map { |attr| send(attr) }
    end

    # see columns_sans_id
    def attribute_values_sans_id
      self.class.columns_sans_id.map { |attr| send(attr) }
    end

    def insert
      last_row_id = DBConnection.insert(<<-SQL, *attribute_values_sans_id)
      INSERT INTO
        #{self.class.table_name}(#{self.class.columns_sans_id.join(',')})
      VALUES
        (#{(["?"] * attribute_values_sans_id.length).join(',')});
    SQL
      self.id = last_row_id
    end

    def update
      DBConnection.execute(<<-SQL, *attribute_values_sans_id, self.id)
      UPDATE
        #{self.class.table_name}
      SET
        #{self.class.columns_sans_id.map { |attr_name| "#{attr_name}=?"}.join(',')}
      WHERE
        id = ?;
    SQL
    end

    def save
      if self.id.nil?
        insert
      else
        update
      end
    end
  end
end
